
import re
def split(s):
    """
    Splits a triple-quoted string by whitespace.

    Also allows comments - anything from '#' to EOL is removed.

    This is used to make it easier to maintain long lists of modules in a simple string, yet also
    have some minimal documentation comments.
    """
    s = s.strip()
    s = re.sub(r'#[^\n]+\n', '', s)
    s = re.sub('[\r\n\t]', '', s)
    a = s.split()

    if __debug__:
        from collections import Counter
        c = Counter(a)
        dups = [key for (key,count) in c.items() if count > 1]
        if dups:
            print('ERROR: Duplicate entries')
            for dup in dups:
                print(' ' + dup)
            import sys
            sys.exit(1)

    return s.split()




import sys, logging, shutil, zipfile, time, os
import marshal
from os.path import dirname, join, exists, abspath, isdir
from io import BytesIO
from py_compile import compile


class ExeBuilder:
    def __init__(self, logger=None, version_strings=None, extra=None, *, filename, version,
                 analysis, subsystem):
        """
        filename
          The fully qualified name of the executable to build.  The directory must already
          exist.

        extra
          An optional sequence of additional files (not Python source) to add to the
          executable.  Each element can be a relative path or a tuple pair containing
          (filename, archivename) where `archivename` is the name to store the file as in the
          executable.
        """
        if subsystem not in ('console', 'windows'):
            raise ValueError('subsystem must be "windows" or "console"')

        assert isdir(dirname(filename))

        self.filename = filename
        self.analysis = analysis
        self.subsystem = subsystem
        self.extra = extra
        self.version = version
        self.version_strings = version_strings

        self.logger = logger or logging.getLogger('kelvin')

    def build(self):
        fqn = self.get_src_exe()
        self.logger.debug('Copying %s to %s', fqn, self.filename)
        shutil.copy(fqn, self.filename)

        # I seem to be getting the wrong file sometimes.
        assert _copied(fqn, self.filename)

        from .resources import AddVersionResource
        AddVersionResource(self.filename, self.version, self.version_strings)

        # Append the Python modules and extra files to the executable.

        fd = open(self.filename, 'ab')
        fd.write(self.zip_files())
        fd.close()

    def zip_files(self):
        """
        Zips Python modules and returns the content.
        """
        buffer = BytesIO()
        zf = zipfile.ZipFile(buffer, mode='w', compression=zipfile.ZIP_DEFLATED)

        timestamp = int(time.time())

        ext = '.pyc'

        tempdir = dirname(self.filename)

        for m in self.analysis.modules:
            # Python creates its own __main__, which keep us from being able to import the real
            # one using that name.  Package the main module under the name __kelvinmain__.

            name = m.__name__ == '__main__' and '__kelvinmain__' or m.__name__

            if m.__file__.endswith('__init__.py'):
                arcname = join(name.replace('.', '/'), '__init__{}'.format(ext))
            else:
                arcname = name.replace('.', '/') + ext

            bytes = self._byte_compile(name, m.__file__, timestamp, tempdir)
            self.logger.log(1, 'storing %s bytes as %s', len(bytes), arcname)

            zf.writestr(arcname, bytes)

        # Now add extra files.  (Need to rearrange the modules here.)

        if self.extra:
            for item in self.extra:
                if isinstance(item, tuple):
                    filename, arcname = item
                else:
                    # The item is a single string which will be used for both the source filename and the archive name.
                    # The name must be a relative path since we don't know now to make the archive name ourselves.

                    if isabs(item):
                        raise Exception('extra file "{}" needs to be a relative path or the relative path in the zip must be provided'.format(item))

                    filename, arcname = item, item

                if not exists(filename):
                    raise Exception('extra file "{}" not found'.format(filename))

                arcname = arcname.replace('\\', '/')

                zf.write(filename, arcname)

        zf.close()

        return buffer.getvalue()

    def get_src_exe(self):
        """
        Find the executable we will copy and modify.
        """
        # When packaged as a binary wheel, the kelvinc.exe and kelvinw.exe files are in a data
        # directory at the root.

        root = dirname(abspath(__file__))

        assert self.subsystem in ('console', 'windows')
        filename = 'kelvin{}.exe'.format(self.subsystem[0])

        path = join(root, 'data', filename)
        if exists(path):
            return path

        # If we are working on Kelvin, such as using "pip --edit kelvin", grab the latest build
        # from the distutils build directory.

        build = self._find_build_dir()
        if build:
            path = join(build, 'data', filename)
            if exists(path):
                self.logger.debug('exepath=%s', path)
                return path
            self.logger.debug('exe not found: %s', path)

        raise Exception('{} not found.  Run setup.py build or install'.format(filename))

    def _find_build_dir(self):
        # distutils really sucks.  I need to know where the executables were built, but other
        # than creating a fake distutils.dist instance, I don't see how we can find the
        # directory.  We'll copy the same logic and hope it doesn't change much.
        #
        # This is only used for the development of Kelvin, so I'm not worried about it handling
        # every possibility.

        try:
            from distutils import util
        except:
            return None

        plat_name = util.get_platform()
        lib = "lib.%s-%d.%d" % (plat_name, *sys.version_info[:2])

        fqn = abspath(join(dirname(__file__), '..', 'build', lib, 'kelvin'))
        if not exists(fqn):
            self.logger.debug('libdir does not exist %s', fqn)
            return None

        self.logger.debug('libdir=%s', fqn)
        return fqn


    def _byte_compile(self, name, path, timestamp, tempdir):
        """
        Byte compiles a Python file and returns the bytes.

        path
        The path to the original source
        """

        # As of Python 3.6, there isn't an API to do this without writing to a physical file by
        # giving it a pathname.  We'll use a temp file - this is going to be kind of slow.
        #
        # I can't use any of the secure functions in tempfile since they all create the file, but
        # py_compile doesn't allow that.  We'll just make something in the dist directory and then
        # delete it.

        cfile = join(tempdir, 'temp.pyc')
        if exists(cfile):
            os.unlink(cfile)

        self.logger.log(1, 'compile: %s --> %s', path, cfile)

        try:
            compile(path, cfile=cfile, optimize=1)
        except Exception as details:
            raise Exception("compiling '%s' failed\n    %s: %s" % (path, details.__class__.__name__, details))

        # If a compilation error occurred, the error will have been written to the screen.
        if not exists(cfile):
            sys.exit('A compilation error occurred')

        b = open(cfile, 'rb').read()
        os.unlink(cfile)
        return b


def _copied(src, dst):
    """
    A debugging function used in an assertion to make sure the source file was successfully
    copied to the destination.
    """
    ss = os.stat(src)
    ds = os.stat(dst)

    if ss.st_size != ds.st_size:
        raise AssertionError('src=%s (%s) dst=%s (%s)' % (src, ss.st_size, dst, ds.st_size))

    return True




# Compatability items for Python 2.7

import sys
from cStringIO import StringIO as BytesIO

# Python 2.7 support is temporary right now.  There is no optimize keyword to the compile
# function.
if sys.version_info < (2, 7):
    raise Exception('Kelvin requires Python 2.7 or later')

def toucs2(value):
    """
    Converts str or unicode objects to a byte array encoded as UTF16-LE, which is the closest
    to a simple UCS2 Python has.
    """
    if type(value) is str:
        value = unicode(value) # , 'unicode_escape')
    return value.encode('utf_16_le')


def tounicode(value):
    if type(value) is str:
        value = unicode(value) # , 'unicode_escape')
    return value
    
def istext(value):
    return isinstance(value, (str, unicode))

def isunicode(value):
    return isinstance(value, unicode)

def joinbytes(l):
    return ''.join(l)

padding = '\0' * 3



import sys, os, shutil
from os.path import isdir, join, splitext, basename
import logging

class Builder:
    """
    Builds the executable and copies files to the distribution directory.
    """
    def __init__(self, include=None, exclude=None, path=None, report=None, logger=None,
                 package_paths=None, subsystem='console', filename=None, expand=None,
                 extra=None, version_strings=None,
                 *, script, dist, version):
        """
        See the README.rst file for a description of the parameters.
        """
        if subsystem not in ('console', 'windows'):
            raise ValueError('subsystem must be "windows" or "console"')

        self.script  = script
        self.path    = path
        self.include = include
        self.exclude = exclude
        self.report  = report
        self.package_paths = package_paths
        self.logger = logger or logging.getLogger('kelvin')
        self.subsystem = subsystem
        self.expand = expand

        if filename:
            if os.path in filename:
                raise ValueError("filename must not contain a path")
        else:
            filename = splitext(basename(script))[0] + '.exe'
        self.filename = filename

        self.dist = dist
        self.extra = extra
        self.version = version
        self.version_strings = version_strings

    def build(self):
        """
        Build the distribution into the dist directory.

        This will create the dist directory if necessary or delete the contents if it already
        exists.
        """
        self.logger.info('Building to %s', self.dist)
        self.prepare_dist_dir()

        from .analyze import Analyzer
        from .dlls import DllFinder

        analyzer = Analyzer(script=self.script, report=self.report, path=self.path, package_paths=self.package_paths,
                            include=self.include, exclude=self.exclude, expand=self.expand)
        analysis = analyzer.analyze()

        if analysis.missing:
            print('Missing')
            for name in sorted(analysis.missing):
                print('-', name)
            sys.exit(1)

        from pprint import pprint

        dllfinder = DllFinder(analysis.extensions)
        dlls = dllfinder.find_dlls()

        if dlls.missing:
            print('Missing DLLs')
            for name in sorted(dlls.missing):
                print('-', name)
            sys.exit(1)

        from .build_exe import ExeBuilder
        filename = join(self.dist, self.filename)
        exebuilder = ExeBuilder(filename=filename, analysis=analysis, subsystem=self.subsystem,
                                version=self.version, version_strings=self.version_strings,
                                extra=self.extra)
        exebuilder.build()

        self.copy_dlls(dlls)

    def prepare_dist_dir(self):
        """
        Make sure the dist directory exists and is empty.
        """
        if isdir(self.dist):
            # Windows won't allow us to delete a directory if it is a terminal's current
            # directory.  To make it easy to test, just clear the directory instead of deleting
            # it.
            self.logger.debug('emptying existing dist directory')
            for filename in os.listdir(self.dist):
                fqn = join(self.dist, filename)
                if isdir(fqn):
                    shutil.rmtree(fqn)
                else:
                    os.unlink(fqn)
        else:
            self.logger.debug('creating dist directory')
            os.makedirs(self.dist)

        # I've been having problems with corrupted executables.  Perhaps the directory
        # is not getting cleaned out properly?
        assert not list(os.listdir(self.dist))


    def copy_dlls(self, dlls):
        for f in dlls.found:
            self.logger.log(1, '%s --> %s', f, self.dist)
            shutil.copy(f, self.dist)




import sys, os, re, logging
from os.path import dirname, basename, isabs, join, abspath, exists
from collections import namedtuple
from .imageutils import find_dependencies

DllList = namedtuple('DllList', 'found missing')

WHITELIST = re.compile(r'(api-ms|vcruntime|ucrtbase)')

class DllFinder:
    def __init__(self, extensions, exclude=None, logger=None):
        self.extensions = extensions
        self.exclude = exclude
        self.logger = logger or logging.getLogger('kelvin.dlls')

        self.windir = os.environ['WINDIR'].lower()

    def find_dlls(self):
        """
        Finds the DLLs imported by (needed by) the extensions.

        extensions
          A list of modulefinder.Module objects for the Python extension DLLs.
        """
        path = [ dirname(sys.executable) ]
        path.extend(os.environ['PATH'].split(';'))
        path.extend(sys.path)
        path = ';'.join(path)

        self.logger.log(1, 'DLL search path: %s', path)

        map_module_to_path = {}

        images = [sys.executable]  # picks up python DLL

        fqn = _find_file('ucrtbase.dll', path)
        if not fqn:
            logger.error('Unable to find ucrtbase.dll.  Not all DLLs may be found!')
        else:
            images.append(fqn)
            map_module_to_path[basename(fqn)] = fqn

        # Include any extensions, which are DLLs, that the analyzer found we need.
        # Include in the list of DLLs to copy (map_module_to_path) and also in the
        # list of images to look for dependencies for (images).

        for item in self.extensions:
            images.append(item.__file__)
            map_module_to_path[basename(item.__file__)] = item.__file__

        seen = set()
        # The images we've already processed, as fully qualified paths.  The same as the values
        # in map_module_to_path.

        while images:
            image = images.pop(0)
            assert isabs(image), 'Not absolute path: {}'.format(image)

            dep = find_dependencies(image, path)
            self.logger.log(1, 'Dependencies: image=%s deps=%r', image, dep)

            for filename, fqn in dep.items():
                if fqn in seen:
                    continue
                seen.add(fqn)

                if self._exclude(fqn):
                    self.logger.debug('Excluding DLL %s', fqn)
                    continue

                images.append(fqn)
                map_module_to_path[filename] = fqn

        # Separate the missing and found.

        found = list(map_module_to_path.values())

        # TODO: Do we know when a DLL is missing now?
        # missing = [ module for (module, fqn) in map_module_to_path.items() if module not in exclude and fqn is None ]
        missing = []

        return DllList(found, missing)

    def _exclude(self, fqn):
        """
        Returns True if this DLL should be excluded.  We need to exclude during processing
        so we don't pick up dependencies of an excluded DLL.
        """
        fqn = fqn.lower()

        filename = basename(fqn)

        # Always include pyd files unless they have been explicitly excluded.  (I don't think
        # this can happen, but there is no since in making it impossible.)
        if self.exclude and filename.endswith('.pyd'):
            return filename[:-4] in self.exclude

        if WHITELIST.match(filename):
            return False

        if fqn.startswith(self.windir):
            return True

        return False


def _find_file(filename, path):
    """
    Searches for the given filename in the path string.  If found, returns the absolute path.
    Otherwise returns None.
    """
    for dir in path.split(';'):
        fqn = join(abspath(dir), filename)
        if exists(fqn):
            return fqn
    return None




# Compatability items for Python 3.2

import sys
from io import BytesIO

if sys.version_info < (3, 6):
    raise Exception('Kelvin requires Python 3.6 or later')

def toucs2(value):
    """
    Converts unicode objects to a bytes object encoded as UTF16-LE, which is the closest to a
    simple UCS2 Python has.
    """
    return value.encode('utf_16_le')


def tounicode(value):
    return value

def istext(value):
    return isinstance(value, str)

def isunicode(value):
    return isinstance(value, str)


def joinbytes(l):
    return b''.join(l)

padding = b'\0' * 3




from ctypes import *
from os.path import exists, basename, join, dirname, isabs
from distutils.errors import *

PSTR      = c_char_p
DWORD     = c_uint
ULONG_PTR = POINTER(c_ulong)

def find_dependencies(image, path):
    """
    Returns DLLs that `image` is dependent on as a mapping from lowercased DLL file name to its
    fully qualified path.  If a DLL is not found on the path, it maps to None.

    image
      Fully qualified path to an executable or DLL.

    path
      A Windows path string to search.
    """
    ih = windll.Imagehlp
    bi = ih.BindImageEx
    bi.argtypes = [ DWORD, PSTR, PSTR, PSTR, IMAGEHLP_STATUS_ROUTINE ]

    GetLastError = windll.kernel32.GetLastError

    BIND_NO_BOUND_IMPORTS = 0x00000001
    BIND_NO_UPDATE        = 0x00000002
    BIND_ALL_IMAGES       = 0x00000004

    assert exists(image), image

    # Strangely, Imagehlp only supports ASCII filenames.
    imageA = image.encode('UTF8')
    pathA  = path.encode('UTF8')

    global map_module_to_path
    map_module_to_path = {}

    callback = IMAGEHLP_STATUS_ROUTINE(StatusRoutine)
    b = bi(BIND_NO_BOUND_IMPORTS | BIND_ALL_IMAGES | BIND_NO_UPDATE, imageA, pathA, None, callback)
    if b == 0:
        raise DistutilsInternalError('BindImageEx({}) failed with error {}'.format(image, GetLastError()))

    return map_module_to_path


BindOutOfMemory           = 1
BindRvaToVaFailed         = 2
BindNoRoomInImage         = 3
BindImportModuleFailed    = 4
BindImportProcedureFailed = 5
BindImportModule          = 6
BindImportProcedure       = 7
BindForwarder             = 8
BindForwarderNOT          = 9
BindImageModified         = 10
BindExpandFileHeaders     = 11
BindImageComplete         = 12
BindMismatchedSymbols     = 13
BindSymbolsNotUpdated     = 14
BindImportProcedure32     = 15
BindImportProcedure64     = 16
BindForwarder32           = 17
BindForwarder64           = 18
BindForwarderNOT32        = 19
BindForwarderNOT64        = 20

seen_modules = set() # module names with no path
seen_fqn     = [] # fully qualified names
bad_modules = set() # Modules that could not be loaded

map_module_to_path = {}
# If a module is required but not found, the value will be None.

IMAGEHLP_STATUS_ROUTINE = WINFUNCTYPE(c_int, DWORD, PSTR, PSTR, ULONG_PTR, ULONG_PTR)

def StatusRoutine(reason, imageName, dllName, va, param):
    dllName = dllName.decode('ascii')

    if isabs(dllName):
        module = basename(dllName).lower()
        map_module_to_path[module] = dllName
        return 1

    else:
        # dllName is just a module name.  Since it might be passed in as a fully-qualified path
        # later, just store it for now and we'll resolve them to full paths elsewhere.
        module = dllName.lower()
        if module not in map_module_to_path:
            map_module_to_path[module] = None

    return 1


if __name__ == '__main__':
    imagename  = "C:\\bin\\python32\\python.exe"
    searchpath = "C:\\bin\\python32;c:\\bin\\java6\\bin;C:\\Windows\\system32;C:\\Windows;C:\\Windows\\System32\\Wbem;C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\;C:\\Program Files\\Intel\\WiFi\\bin\\;C:\\Program Files\\Common Files\\Intel\\WirelessCommon\\;C:\\Program Files (x86)\\Common Files\\Roxio Shared\\OEM\\DLLShared\\;C:\\Program Files (x86)\\Common Files\\Roxio Shared\\OEM\\DLLShared\\;C:\\Program Files (x86)\\Common Files\\Roxio Shared\\OEM\\12.0\\DLLShared\\;C:\\Program Files (x86)\\Roxio\\OEM\\AudioCore\\;C:\\Program Files (x86)\\Microsoft SQL Server\\100\\Tools\\Binn\\;C:\\Program Files\\Microsoft SQL Server\\100\\Tools\\Binn\\;C:\\Program Files\\Microsoft SQL Server\\100\\DTS\\Binn\\;C:\\Program Files (x86)\\Microsoft SQL Server\\100\\Tools\\Binn\\VSShell\\Common7\\IDE\\;C:\\Program Files (x86)\\Microsoft SQL Server\\100\\DTS\\Binn\\;C:\\bin\\mercurial\\;C:\\bin\\graphviz\\bin;C:\\PROGRA~2\\IBM\\GSK7_64\\bin;C:\\PROGRA~2\\IBM\\GSK7_64\\lib64;C:\\Program Files\\IBM\\GSK7_64\\bin;C:\\Program Files\\IBM\\GSK7_64\\lib64;C:\\PROGRA~2\\ibm\\gsk7\\bin;C:\\PROGRA~2\\ibm\\gsk7\\lib;C:\\Program Files (x86)\\ibm\\gsk7\\bin;C:\\Program Files (x86)\\ibm\\gsk7\\lib;c:\\bin\\git\\cmd;c:\\bin\\git\\bin;C:\\bin\\svn\\bin;C:\\Program Files (x86)\\QuickTime\\QTSystem\\;c:\\bin\\utils;c:\\bin\\git\\bin;c:\\bin\\python27;c:\\bin\\emacs\\bin;c:\\bin\\python27\\Scripts;C:\\dev\\kelvin\\kelvin;C:\\dev\\kelvin\\tmp;C:\\Windows\\system32\\python32.zip;C:\\bin\\python32\\DLLs;C:\\bin\\python32\\lib;C:\\bin\\python32;C:\\bin\\python32\\lib\\site-packages"
    found, missing = find_dependencies(imagename, searchpath)
    print('found:', found)
    print('missing:', missing)




# The first Python module run by the executable.  It looks for a module named __kelvinmain__
# (the user's main module) and runs it with the name '__main__'.

import sys
try:
    from pkgutil import get_loader # Python 3
except ImportError:
    from imp import get_loader     # Python 2

def run_main():
    """
    The Kelvin code that runs the application's startup module as __main__.
    """
    # This seems kind of messy.  It is required because Python automatically creates a __main__
    # module in Py_Initialize and we have to run a different module, but fool it into thinking
    # it is __main__.  We do this by running the module, but passing it the dictionary from
    # __main__.

    mod_name = '__kelvinmain__'

    loader   = get_loader(mod_name)
    code     = loader.get_code(mod_name)
    filename = loader.get_filename(mod_name)  # Python 2.7+.  Earlier versions may have _get_filename.

    # # Python 3 has zipmodule.get_filename, but it is unofficial and _get_filename in Python 2.
    # for fname in ('get_filename', '_get_filename'):
    #     func = getattr(loader, fname, None)
    #     if func:
    #         filename = func(mod_name)
    #         break

    globals = sys.modules["__main__"].__dict__

    globals.update(
        __name__    = '__main__',
        __file__    = filename,
        __loader__  = loader,
        __cached__  = None,
        __package__ = None)

    exec(code, globals)



#!/usr/bin/python

from __future__ import print_function

"A distutils extension to freeze Python 3 scripts into Windows executables."

from setuptools import setup

import sys, os, re
from distutils.core import Command
from distutils.errors import DistutilsError
from distutils.extension import Extension
from os.path import abspath, dirname, exists, join
from distutils.util import get_platform

from build_exe import *


def main():
    minver = (3,6)
    if (sys.version_info.major, sys.version_info.minor) < minver:
        raise DistutilsError('Kelvin requires Python %s.%s or later' % minver)

    version = get_version()

    files = [ abspath(join('src', f)) for f in os.listdir('src') if f.endswith('.cpp') ]

    settings = { 'define_macros' : [('KELVIN_VERSION', version)],
                 'libraries' : ['Imagehlp', 'Shell32'] }

    # We need to set a build_lib since most of the distutils commands change it depending on
    # whether there are extensions or not.  We don't have extensions, but we do have
    # executables which should be treated the same.
    build_lib = join('build', 'lib.{}-{}'.format(get_platform(), sys.version[0:3]))
    exe_dir   = join(build_lib, 'kelvin', 'data')

    setup(
        distclass = Distribution2,
        name             = 'kelvin',
        version          = version,
        description      = 'Freezes Python scripts into Windows apps',
        long_description = __doc__,
        url              = 'http://www.gitlab.com/mkleehammer/kelvin',
        author           = 'Michael Kleehammer',
        author_email     = 'michael@kleehammer.com',

        packages    = ['kelvin'],

        exe_modules = [
            Executable('kelvinc', 'console', files, **settings),
            Executable('kelvinw', 'windows', files, **settings)
            ],

        options = { 'build'     : { 'build_lib': build_lib },
                    'build_exe' : { 'build_lib': exe_dir   },
                    'clean'     : { 'build_lib': build_lib },
                    'install'   : { 'build_lib': build_lib } },

        classifiers = ['Development Status :: 3 - Alpha',
                       'Intended Audience :: Developers',
                       'License :: OSI Approved :: MIT License',
                       'Operating System :: Microsoft :: Windows',
                       'Programming Language :: Python',
                       'Programming Language :: Python :: 3',
                       ],

        cmdclass = { 'build_exe': BuildExeCommand }
        )


def get_version():
    """
    Returns the version as a string.

    1. If in an unzipped source directory (from setup.py sdist), returns the version from the PKG-INFO file.
    2. If in a git repository, uses the latest tag from git describe.
    3. This should probably be an error, but it will return 0.1b1.
    """
    return _get_version_pkginfo() or _get_version_git() or '0.1b1'


def _get_version_pkginfo():
    filename = join(dirname(abspath(__file__)), 'PKG-INFO')
    if exists(filename):
        re_ver = re.compile(r'^Version: \s+ (\d.*)', re.VERBOSE)
        for line in open(filename):
            match = re_ver.search(line)
            if match:
                return match.group(1).strip()
    return None


def _get_version_git():
    n, result = getoutput('git describe --tags')
    if n:
        print('WARNING: git describe failed with: %s %s' % (n, result))
        return None

    match = re.match(r'(\d+).(\d+).(\d+)', result)
    if not match:
        return None

    major = int(match.group(1))
    minor = int(match.group(2))
    patch = int(match.group(3))

    modified = bool(getoutput('git --no-pager diff --quiet')[0]) # working directory
    if not modified:
        modified = bool(getoutput('git --no-pager diff --quiet --cached')[0]) # index

    version = '{}.{}.{}'.format(major, minor, patch)

    if modified:
        version += '.dev1'

    return version


def getoutput(cmd):
    pipe = os.popen(cmd, 'r')
    text   = pipe.read().rstrip('\n')
    status = pipe.close() or 0
    return status, text


if __name__ == '__main__':
    main()




import os, sys, logging
import modulefinder
from modulefinder import ReplacePackage
from os.path import join, exists, dirname, abspath, splitext
from collections import namedtuple

from .utils import split

DEFAULT_EXCLUDES = split("""
    _dummy_thread _dummy_threading _emx_link _gestalt _posixsubprocess ce doctest fcntl grp
    java.lang org.python.core os2 os2emxpath posix pwd pydoc readline riscos
    riscosenviron riscospath rourl2path sitecustomize termios unittest usercustomize vms_lib
    win32api win32con win32pipe

    _frozen_importlib _frozen_importlib_external

    # IIRC, OS-specific modules are loaded as os.path, so there isn't an actual file os/path.
    os.path

    # Module finder believes the following are modules and cannot import them so they are excluded by default.

    ctypes._SimpleCData
    xml.dom.EMPTY_NAMESPACE xml.dom.EMPTY_PREFIX xml.dom.Node xml.dom.XMLNS_NAMESPACE xml.dom.XML_NAMESPACE
    """)


Analysis = namedtuple('Analysis', 'modules extensions missing')


class Analyzer:
    """
    Analyzes the source to collect all of the dependencies.
    """
    def __init__(self, include=None, exclude=None, path=None, report=None, logger=None,
                 package_paths=None, expand=None, *, script):
        """
        expand
          A list of package names for which *all* files in the package should be included, not
          just those that module finder has determined are used.

          The special name '.' can be used to represent the directory the main script is in.
        """
        self.script  = script
        self.path    = path
        self.include = include
        self.exclude = DEFAULT_EXCLUDES
        self.report  = report
        self.logger  = logger or logging.getLogger('analyze')

        self.package_paths = package_paths
        self.expand = expand

        if exclude:
            self.exclude += exclude

    def get_finder(self):
        """
        Returns a ModuleFinder instance configured for this application.
        """
        path = self.path
        path += sys.path

        if self.package_paths:
            for pkg, dirs in self.package_paths.items():
                for dir in dirs:
                    modulefinder.AddPackagePath(pkg, dir)

        ReplacePackage('_xmlplus', 'xml')

        finder = modulefinder.ModuleFinder(
            path=path,
            excludes=self.exclude,
            debug=bool(self.report)
        )
        return finder

    def analyze(self):
        finder = self.get_finder()
        finder.run_script(self.script)

        if self.include:
            finder.import_hook('api')
            for module in self.include:
                finder.import_hook(module)

        # Add __kelvin__, our startup module.

        fqn = join(dirname(abspath(__file__)), '__kelvin__.py')
        assert exists(fqn), 'Not found: %r' % fqn
        finder.load_file(fqn)

        self.find_submodules(finder, 'encodings')

        if self.report:
            finder.report()

        return self.parse_mf_results(finder)

    def parse_mf_results(self, finder):
        modules    = []
        extensions = []

        for item in finder.modules.values():
            src = item.__file__
            if not src:
                # Part of Python
                continue

            ext = splitext(src)[1]

            if ext in ['.py', '.pyc', '.pyo']:
                modules.append(item)
                continue

            if ext == '.pyd':
                extensions.append(item)
                continue

            raise Exception('Do not know how to handle {!r}'.format(src))

        missing, _maybe_ = finder.any_missing_maybe()

        # missing.extend(finder.badmodules.keys())

        # I don't know why (yet), but finder puts things into badmodules even though they are
        # supposed to be excluded.  We'll manually remove them until we get to the bottom of
        # this.
        missing = set(missing) - set(self.exclude)

        return Analysis(modules, extensions, missing)

    def find_submodules(self, finder, name):
        self.logger.debug('Importing all submodules for %s', name)
        finder.import_hook(name)
        module = finder.modules[name]
        for submodule in finder.find_all_submodules(module):
            finder.import_hook(name + '.' + submodule)




# Python functions for dealing with Windows resources.

# The encodings should be UCS2 but Python 3.6 does not seem to have it.  I'm going to use
# UTF-16LE, so don't use anything that wouldn't be compatible.

# http://blogs.msdn.com/b/oldnewthing/archive/2006/12/20/1332035.aspx
# http://blogs.msdn.com/b/oldnewthing/archive/2006/12/21/1340571.aspx
#
# "Version resources can be viewed as a serialized tree structure. Each node of the tree has a
#  name and associated data (either binary or text), and each node can have zero or more child
#  nodes."

from ctypes import *
import sys, struct, re
from os.path import exists

RT_VERSION  = 16
RT_MANIFEST = 24

VOS_NT_WINDOWS32   = 0x00040004
VS_FF_SPECIALBUILD = 0x00000020
VS_FF_PRERELEASE   = 0x00000002
VFT_APP            = 0x00000001

WORD  = c_short
DWORD = c_long


def AddVersionResource(exe, version, version_strings):
    """
    Adds a version resource to the given executable, constructed from `version_strings`.

    version
      The executable version as a string.  It can only contain numbers and periods and can have
      up to 4 parts.

    version_strings
      Maps from language identifier (int) to a dictionary of key / value pairs that are the
      version strings for that language.  Use 0x0409 for English.
    """
    # Parse version numbers.  We'll take any integers in the order they were found.
    # We need 4 parts total, so fill with 0s and truncate.

    parts = [ int(n) for n in re.findall('\d+', version) ]
    parts = (parts + [ 0, 0, 0, 0 ])[:4]

    ffi = VS_FIXEDFILEINFO()
    ffi.dwSignature        = 0xFEEF04BD
    ffi.dwStrucVersion     = 0x00010000
    ffi.dwFileVersionMS    = (parts[0] << 16) | parts[1]
    ffi.dwFileVersionLS    = (parts[2] << 16) | parts[3]
    ffi.dwProductVersionMS = (parts[0] << 16) | parts[1]
    ffi.dwProductVersionLS = (parts[2] << 16) | parts[3]
    ffi.dwFileFlagsMask    = 0x3F
    ffi.dwFileFlags        = 0x00
    ffi.dwFileOS           = VOS_NT_WINDOWS32
    ffi.dwFileType         = VFT_APP
    
    if version_strings:
        for lang, strings in version_strings.items():
            if 'SpecialBuild' in strings:
                ffi.dwFileFlags |= VS_FF_SPECIALBUILD

    root = VersionNode('VS_VERSION_INFO', ffi)

    sfi = VersionNode('StringFileInfo', None)
    root.children.append(sfi)

    if version_strings:
        for lang, strings in version_strings.items():
            lang = VersionNode('{:04X}{:04X}'.format(lang, 0x04B0), None)
            sfi.children.append(lang)

            for name, value in strings.items():
                lang.children.append(VersionNode(name, (value + '\0').encode('utf_16_le')))

        vfi = VersionNode('VarFileInfo', None)
        root.children.append(vfi)

        words = []
        for lang in version_strings:
            words.append(lang)
            words.append(0x04B0)

        translation = struct.pack('H' * len(words), *words)
        vfi.children.append(VersionNode('Translation', translation))

    UpdateFileVersion(exe, root.tobytes())


class VersionNode:
    def __init__(self, name, value):
        """
        name
          These are str objects in both versions since we cannot use a leading 'u' to force
          them to unicode.  (Python 3.2 does not allow it.)  This is not a problem since it is
          required to be Unicode value when stored, so we can always convert it.

          In Python 2.7, this will be an str object since they are hardcoded in this file and

        value

          Both Unicode values and binary values are stored in the resource, so more care must
          be taken with this parameter.

          In both versions, a ctypes.Structure (used for VS_FIXEDFILEINFO) will be converted to
          a binary format.

          In Python 3, str values are used for Unicode strings and bytes for binary values.  In
          Python 2, unicode values are used for Unicode strings and str objects for binary
          values.
        """
        self.name = name

        # Set wType to string (1) if there is no data to match StringFileInfo.  I'm not sure if it matters.
        self.wType = (value is None or isinstance(value, str)) and 1 or 0

        if isinstance(value, Structure):
            value = string_at(addressof(value), sizeof(value))
        elif isinstance(value, str):
            # We need to write the NULL terminator and it needs to be included in the length.
            value = (value + '\0').encode('utf_16_le')

        self.value = value
        self.children = []


    def tobytes(self):
        """
        Returns this node encoded as an unpadded VERSIONNODE.
        """
        # Not clear in Microsoft's documentation, but if the value is in a String resource
        # block, the length is in *characters*, not bytes.  This is a very bad design.
        #
        # At this time, the only string values supported are in String resource blocks, so
        # anytime wType is 1 we'll set the length to characters.

        if self.value is not None:
            valuelen = len(self.value)
            if self.wType == 1:
                valuelen /= 2
        else:
            valuelen = 0

        parts = [
            '--', # placeholder for cbNode
            struct.pack('HH', valuelen, self.wType),
            (self.name + '\0').encode('utf_16_le')
            ]

        cbNode = sum(len(part) for part in parts)

        if self.value is not None:
            cbNode = self.add_aligned(parts, cbNode, self.value)

        for child in self.children:
            cbNode = self.add_aligned(parts, cbNode, child.tobytes())

        parts[0] = struct.pack('H', cbNode)

        return b''.join(parts)


    def add_aligned(self, parts, length, value):
        """
        Adds `value` to the parts list and returns the updated byte length of items in the list.

        If the current length of items in the list (`length`) is not aligned on a 32-bit
        boundary, padding will be inserted before the value.
        """
        if not value:
            return length

        r = length % 4
        if r:
            padding = (b'\0' * 3)[:(4-r)]
            parts.append(padding)
            length += len(padding)

        parts.append(value)
        length += len(value)

        return length



class VS_FIXEDFILEINFO(Structure):
    _fields_ = [
        ("dwSignature",        DWORD),
        ("dwStrucVersion",     DWORD),
        ("dwFileVersionMS",    DWORD),
        ("dwFileVersionLS",    DWORD),
        ("dwProductVersionMS", DWORD),
        ("dwProductVersionLS", DWORD),
        ("dwFileFlagsMask",    DWORD),
        ("dwFileFlags",        DWORD),
        ("dwFileOS",           DWORD),
        ("dwFileType",         DWORD),
        ("dwFileSubtype",      DWORD),
        ("dwFileDateMS",       DWORD),
        ("dwFileDateLS",       DWORD) ]


def dump(a):
    offset = 0
    while a:
        line, a = a[:16], a[16:]

        parts = ['%04x ' % offset]

        parts.append(' '.join('%02X' % ord(b) for b in line[:8]))
        parts.append('-')
        parts.append(' '.join('%02X' % ord(b) for b in line[8:]))
        parts.append(' ')
        parts.append(''.join((31 <= ord(b) < 128) and b or '.' for b in line))
        print(''.join(parts))

        offset += 16

LANG_NEUTRAL    = 0x00
SUBLANG_NEUTRAL = 0x00

def MAKELANGID(p, s):
    return int(((s & 0xFF) << 10) | (p & 0xFF))

def UpdateFileVersion(filename, version):
    kernel32 = windll.kernel32
    BeginUpdateResource = kernel32.BeginUpdateResourceW
    UpdateResource      = kernel32.UpdateResourceW
    EndUpdateResource   = kernel32.EndUpdateResourceW

    if not exists(filename):
        raise Exception('file does not exist: %s' % filename)

    h = BeginUpdateResource(filename, 0)
    if not h:
        raise WinError()

    if not UpdateResource(h, RT_VERSION, 1, MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL), version, len(version)):
        raise WinError()

    if not EndUpdateResource(h, 0):
        raise WinError()
